module Main (main) where

import qualified LLVM.Extra.ScalarOrVector as SoV
import qualified LLVM.Extra.Arithmetic as A
import LLVM.Extra.Control (arrayLoop, )

import qualified LLVM.Core as LLVM
import LLVM.Core (Value, valueOf, Function, CodeGenModule, CodeGenFunction, )
import LLVM.ExecutionEngine (simpleFunction, )

import qualified Foreign.Storable as St
import Foreign.Marshal.Array (allocaArray, )
import Foreign.Ptr (Ptr, plusPtr)

import qualified System.IO as IO

import Data.Word (Word32, )
import Data.List
import Control.Monad (liftM2, void, forM_)


--------------------------------------------------------------------------------

type Note = Int
type Freq = Float

noteFreq :: Note -> Freq
noteFreq n = 2 ** (fromIntegral n / 12) * (441 / 44100.00)

--------------------------------------------------------------------------------

waveSin :: Value Float -> CodeGenFunction r (Value Float)
waveSin = A.sin

waveSaw :: Value Float -> CodeGenFunction r (Value Float)
waveSaw t =
   A.sub (valueOf 1) =<<
   A.mul (valueOf 2) t

osciSaw ::
   Value Float -> Value Float ->
   CodeGenFunction r (Value Float, Value Float)
osciSaw freq phase =
   liftM2 (,) (waveSin phase) (SoV.incPhase freq phase)


moduleSaw :: CodeGenModule (Function (Word32 -> Float -> Ptr Float -> IO Float))
moduleSaw =
   LLVM.createFunction LLVM.ExternalLinkage $ \ size freq ptr -> do
      s <- arrayLoop size ptr (valueOf 0) $ \ ptri phase0 -> do
         (y, phase1) <- osciSaw freq phase0
         LLVM.store y ptri
         return phase1
      LLVM.ret (s :: Value Float)

writeSawCode :: IO ()
writeSawCode = do
   m <- LLVM.newModule
   void $ LLVM.defineModule m moduleSaw
   LLVM.writeBitcodeToFile "saw.bc" m

renderSaw :: IO ()
renderSaw = do 
   IO.withFile "saw.f32" IO.WriteMode $ \h ->
      let len = 44100 `div` 4
          tones = intersperse 0 $ fmap noteFreq [_c, _d, _e, _f, _g, _g, _A, _A, _A, _A, _g]
      in playFreqs h len tones
         
_a, _h, _c, _d, _e, _f, _g, _A :: Int
_a = 0
_h = 2
_c = 3
_d = 5
_e = 7
_f = 8
_g = 10
_A = 12
         
playFreqs :: IO.Handle -> Int -> [Freq] -> IO ()
playFreqs h len fs = allocaArray (length fs * len) go where
  go :: Ptr Float -> IO ()
  go ptr = do
    fill <- simpleFunction moduleSaw
    forM_ (zip [0..] fs) $ \(i, freq) -> do
      let ptr' = plusPtr ptr $ len * i * St.sizeOf(undefined::Float)
      fill (fromIntegral len) freq ptr'
    IO.hPutBuf h ptr (length fs * len * St.sizeOf(undefined::Float))

main :: IO ()
main = do
   LLVM.initializeNativeTarget
   writeSawCode
   renderSaw
